// DevicePCI.c
//
// Generated by C DriverWizard 3.1.0 (Build 1722)
// Requires DDK Only
// File created on 5/30/2008
//

#include "pch.h"
#include "..\intrface.h"
#ifdef DEVICEPCI_WMI_TRACE
#include "DevicePCI.tmh"
#endif

// global data
DEVICEPCI_DATA g_Data;

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DriverEntry 
//      Installable driver initialization entry point.
//      This entry point is called directly by the I/O system.
//
//  Arguments:
//      IN  DriverObject
//              pointer to the driver object
//
//      IN  RegistryPath
//              pointer to a unicode string representing the path,
//              to driver-specific key in the registry.
//
//  Return Value:
//      Status
//
NTSTATUS DriverEntry(
    IN  PDRIVER_OBJECT  DriverObject,
    IN  PUNICODE_STRING RegistryPath
    )
{
    NTSTATUS    status;

    DevicePCIDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"++");
    DevicePCIDebugPrint(DBG_INIT, DBG_INFO, "Compiled at %s on %s", __TIME__, __DATE__);

#ifdef DBG
//    DbgBreakPoint();
#endif

#ifdef DEVICEPCI_WMI_TRACE 
    WPP_INIT_TRACING(DriverObject, RegistryPath);
#endif

    status = STATUS_SUCCESS;

    RtlZeroMemory(&g_Data, sizeof(DEVICEPCI_DATA));

    // save registry path for wmi
    g_Data.RegistryPath.Length = RegistryPath->Length;
    g_Data.RegistryPath.MaximumLength = RegistryPath->Length + sizeof(UNICODE_NULL);
    g_Data.RegistryPath.Buffer = (PWCHAR)ExAllocatePoolWithTag(
                                            PagedPool,
                                            g_Data.RegistryPath.MaximumLength,
                                            DEVICEPCI_POOL_TAG
                                            );

    if (g_Data.RegistryPath.Buffer == NULL)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

        DevicePCIDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__": Failed to allocate memory for RegistryPath");

        return status;
    }

    RtlCopyUnicodeString(&g_Data.RegistryPath, RegistryPath);

    // detect current operating system
    if (IoIsWdmVersionAvailable(1, 0x30))
    {
        g_Data.WdmVersion = 0x0130;
    }
    else if (IoIsWdmVersionAvailable(1, 0x20))
    {
        g_Data.WdmVersion = 0x0120;
    }
    else if (IoIsWdmVersionAvailable(1, 0x10))
    {
        g_Data.WdmVersion = 0x0110;
    }
    else if (IoIsWdmVersionAvailable(1, 0x05))
    {
        g_Data.WdmVersion = 0x0105;
    }
    else
    {
        g_Data.WdmVersion = 0x0100;
    }

    // setup our dispatch function table in the driver object
    DriverObject->MajorFunction[IRP_MJ_PNP] = DevicePCIPnpDispatch;
    DriverObject->MajorFunction[IRP_MJ_POWER] = DevicePCIPowerDispatch;
    DriverObject->MajorFunction[IRP_MJ_CREATE] = DevicePCICreateDispatch;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = DevicePCICloseDispatch;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DevicePCIDeviceIoControlDispatch;
    DriverObject->MajorFunction[IRP_MJ_READ] = DevicePCIReadDispatch;
    DriverObject->MajorFunction[IRP_MJ_WRITE] = DevicePCIWriteDispatch;
    DriverObject->MajorFunction[IRP_MJ_CLEANUP] = DevicePCICleanupDispatch;
    DriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL] = DevicePCISystemControlDispatch;

    DriverObject->DriverExtension->AddDevice = DevicePCIAddDevice;
    DriverObject->DriverUnload = DevicePCIUnload;

    DevicePCIDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"--. STATUS %x", status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIAddDevice 
//      The PnP manager new device callback
//
//  Arguments:
//      IN  DeviceObject
//              pointer to a device object.
//
//      IN  PhysicalDeviceObject
//              pointer to a device object created by the
//              underlying bus driver.
//
//  Return Value:
//      Status
//
NTSTATUS DevicePCIAddDevice(
    IN PDRIVER_OBJECT   DriverObject,
    IN PDEVICE_OBJECT   PhysicalDeviceObject
    )
{
    NTSTATUS                            status;
    PDEVICE_OBJECT                      deviceObject;
    PDEVICEPCI_DEVICE_EXTENSION   deviceExtension;
    UNICODE_STRING                      ntName = RTL_CONSTANT_STRING(L"\\Device\\DevicePCIDevice");
    UNICODE_STRING                      win32Name = RTL_CONSTANT_STRING(L"\\??\\DevicePCIDevice");

    DevicePCIDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"++: PDO %p", PhysicalDeviceObject);


    // Create our function device object.
    status = IoCreateDevice(
                DriverObject,
                sizeof (DEVICEPCI_DEVICE_EXTENSION),
                &ntName,
                FILE_DEVICE_UNKNOWN,
                FILE_DEVICE_SECURE_OPEN, // Do not use if binary intended for 9x
                FALSE,
                &deviceObject
                );

    if (!NT_SUCCESS(status)) 
    {
        DevicePCIDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__"--. IoCreateDevice STATUS %x", status);

        return status;
    }

    DevicePCIDebugPrint(DBG_INIT, DBG_INFO, __FUNCTION__": Created FDO %p", deviceObject);

    // Initialize the device extension.
    deviceExtension = (PDEVICEPCI_DEVICE_EXTENSION)deviceObject->DeviceExtension;

    // Zero the memory
    RtlZeroMemory(deviceExtension, sizeof(DEVICEPCI_DEVICE_EXTENSION));

    // save the PDO pointer
    deviceExtension->PhysicalDeviceObject = PhysicalDeviceObject;

    // save our device object
    deviceExtension->DeviceObject = deviceObject;

    KeInitializeDpc(&deviceExtension->DevicePCIIsrDpc, DevicePCIDpcForIsr, deviceExtension);

    // set RemoveCount to 1. Transition to zero
    // means IRP_MN_REMOVE_DEVICE was received
    deviceExtension->RemoveCount = 1;

    // Initialize Remove event
    KeInitializeEvent(&deviceExtension->RemoveEvent, NotificationEvent, FALSE);

    // initialize io lock
    DevicePCIInitializeIoLock(&deviceExtension->IoLock, deviceObject);

    deviceExtension->PnpState = PnpStateNotStarted;
    deviceExtension->PreviousPnpState = PnpStateNotStarted;

    // Initialize the device object flags

    // All WDM drivers are supposed to set this flag.  Devices
    // requiring a large amount of current at statup set the 
    // DO_POWER_INRUSH flag instead.  These flags are mutually
    // exclusive.
    if (PhysicalDeviceObject->Flags & DO_POWER_PAGABLE) 
    {
        deviceObject->Flags |= DO_POWER_PAGABLE;
    }

    // This flag sets the buffering method for reads and writes
    // to METHOD_DIRECT.  IOCTLs are handled by IO control codes
    // independent of the value of this flag.
    deviceObject->Flags |= DO_DIRECT_IO;

    // Attach our device to the device stack and get the device object to 
    // which we send down IRPs to.
    deviceExtension->LowerDeviceObject = IoAttachDeviceToDeviceStack(
                                            deviceObject,
                                            PhysicalDeviceObject
                                            );

    if (deviceExtension->LowerDeviceObject == NULL) 
    {
        DevicePCIDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__"--. IoAttachDeviceToDeviceStack failed");

        IoDeleteDevice(deviceObject);
        return STATUS_DEVICE_REMOVED;
    }

    status = IoCreateSymbolicLink(&win32Name, &ntName);
    if (!NT_SUCCESS(status))
    {
        DevicePCIDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__"--. IoCreateSymbolicLink failed STATUS %x", status);

        IoDetachDevice(deviceExtension->LowerDeviceObject);
        IoDeleteDevice(deviceObject);
        return status;
    }

    // increment device instance count
    ++g_Data.InstanceCount;

    // We are all done, so we need to clear the
    // DO_DEVICE_INITIALIZING flag.  This must be our
    // last action in AddDevice
    deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

    DevicePCIDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"--. DevicePCIAddDevice--. STATUS %x", status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIStartDevice
//      Start device handler, sets up resources
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  Irp
//              pointer to the Start IRP
//
//  Return Value:
//      NT status code
//
NTSTATUS DevicePCIStartDevice(
    IN  PDEVICEPCI_DEVICE_EXTENSION    DeviceExtension,
    IN  PIRP                            Irp
    )
{
    NTSTATUS                        status = STATUS_SUCCESS;
    PCM_PARTIAL_RESOURCE_DESCRIPTOR resource;
    PCM_PARTIAL_RESOURCE_DESCRIPTOR resourceTrans;
    PCM_PARTIAL_RESOURCE_LIST       partialResourceList;
    PCM_PARTIAL_RESOURCE_LIST       partialResourceListTranslated;
    PIO_STACK_LOCATION              irpStack;
    POWER_STATE                     powerState;
    ULONG                           ii;
    ULONG                           portCount;

    DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    // Get our current IRP stack location
    irpStack = IoGetCurrentIrpStackLocation(Irp);

    // Do whatever initialization needed when starting the device:
    // gather information about it,  update the registry, etc.
    // At this point, the lower level drivers completed the IRP
    if ((irpStack->Parameters.StartDevice.AllocatedResources != NULL) &&
        (irpStack->Parameters.StartDevice.AllocatedResourcesTranslated != NULL)) 
    {
        // Parameters.StartDevice.AllocatedResources points to a
        // CM_RESOURCE_LIST describing the hardware resources that
        // the PnP Manager assigned to the device. This list contains
        // the resources in raw form. Use the raw resources to program
        // the device.

        partialResourceList =
            &irpStack->Parameters.StartDevice.AllocatedResources->List[0].PartialResourceList;

        resource = &partialResourceList->PartialDescriptors[0];

        // Parameters.StartDevice.AllocatedResourcesTranslated points
        // to a CM_RESOURCE_LIST describing the hardware resources that
        // the PnP Manager assigned to the device. This list contains
        // the resources in translated form. Use the translated resources
        // to connect the interrupt vector, map I/O space, and map memory.

        partialResourceListTranslated =
            &irpStack->Parameters.StartDevice.AllocatedResourcesTranslated->List[0].PartialResourceList;

        resourceTrans = &partialResourceListTranslated->PartialDescriptors[0];

        // reset the total port count
        portCount = 0;

        // search the resources
        for (ii = 0;
             ii < partialResourceList->Count;
             ++ii, ++resource, ++resourceTrans) 
        {

            switch (resource->Type) 
            {
            case CmResourceTypePort:
                DevicePCIDebugPrint(DBG_PNP, DBG_INFO,  "Resource Type Port");
                DevicePCIDebugPrint(DBG_PNP, DBG_INFO, "Port.Start %I64x", resource->u.Port.Start.QuadPart);
                DevicePCIDebugPrint(DBG_PNP, DBG_INFO, "Port.Length %x", resource->u.Port.Length);

                switch (portCount)
                {
                case 0:
                    if (resourceTrans->Type == CmResourceTypePort)
                    {
                        DeviceExtension->bResource_portIsIo = TRUE;

                        DeviceExtension->Resource_portLength = resourceTrans->u.Port.Length;
                        DeviceExtension->bResource_portMapped = (resourceTrans->Flags & CM_RESOURCE_PORT_IO) == 0;
                        if (DeviceExtension->bResource_portMapped)
                        {
                            DeviceExtension->Resource_portBase = (PUCHAR)MmMapIoSpace(
                                                                    resourceTrans->u.Port.Start, 
                                                                    resourceTrans->u.Port.Length, 
                                                                    MmNonCached
                                                                    );
                        }
                        else
                        {
                            DeviceExtension->Resource_portBase = (PUCHAR)resourceTrans->u.Port.Start.QuadPart;
                        }
                    }
                    else
                    {
                        ASSERT(resourceTrans->Type == CmResourceTypeMemory);

                        DeviceExtension->bResource_portIsIo = FALSE;
                        DeviceExtension->bResource_portMapped = TRUE;

                        DeviceExtension->Resource_portLength = resourceTrans->u.Memory.Length;
                        DeviceExtension->Resource_portBase = (PUCHAR)MmMapIoSpace(
                                                            resourceTrans->u.Memory.Start, 
                                                            resourceTrans->u.Memory.Length, 
                                                            MmNonCached
                                                            );
                    }
                    break;
                default:
                    DevicePCIDebugPrint(DBG_PNP, DBG_INFO,  "Extra Port");
                    break;
                }

                ++portCount;
                break;
            case CmResourceTypeMemory:
                DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, "Resource Type Memory");
                DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, "Memory.Start %I64x", resource->u.Memory.Start.QuadPart);
                DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, "Memory.Length %x", resource->u.Memory.Length);
                break;
            case CmResourceTypeInterrupt:
                DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, "Resource Type Interrupt");
                DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, "Interrupt.Level %x", resourceTrans->u.Interrupt.Level);
                DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, "Interrupt.Vector %x", resourceTrans->u.Interrupt.Vector);
                DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, "Interrupt.Affinity %x\n", resourceTrans->u.Interrupt.Affinity);

                DeviceExtension->DevicePCIInterruptVector = resourceTrans->u.Interrupt.Vector;
                DeviceExtension->DevicePCIInterruptLevel = (KIRQL)resourceTrans->u.Interrupt.Level;
                DeviceExtension->DevicePCIInterruptAffinity = resourceTrans->u.Interrupt.Affinity;
                DeviceExtension->bDevicePCIInterruptShared = resourceTrans->ShareDisposition == CmResourceShareShared;
                DeviceExtension->DevicePCIInterruptMode = 
                    (resourceTrans->Flags == CM_RESOURCE_INTERRUPT_LATCHED) ? Latched : LevelSensitive;
                break;
            default:
                DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, "Unknown Resource Type %x", resourceTrans->Type);
                break;
            }
        }
    }

    //*****************************************************************
    //*****************************************************************
    // TODO: check that all required resources have been allocated
    //*****************************************************************
    //*****************************************************************

    if (portCount < 1)
    {
        return STATUS_DEVICE_CONFIGURATION_ERROR;
    }
    
    if (DeviceExtension->Resource_portBase == NULL)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    if (DeviceExtension->DevicePCIInterruptVector == 0)
    {
        return STATUS_DEVICE_CONFIGURATION_ERROR;
    }
    //*****************************************************************
    //*****************************************************************
    // TODO: allocate and initialize dma and shared memory resources
    //*****************************************************************
    //*****************************************************************
    //*****************************************************************
    //*****************************************************************
    // TODO: Disable interrupts
    //*****************************************************************
    //*****************************************************************

    // Connect our ISR to the interrupt resource
    status = IoConnectInterrupt(
                &DeviceExtension->DevicePCIInterruptObject,
                DevicePCIIsr,
                DeviceExtension,
                NULL,
                DeviceExtension->DevicePCIInterruptVector,
                DeviceExtension->DevicePCIInterruptLevel,
                DeviceExtension->DevicePCIInterruptLevel,
                DeviceExtension->DevicePCIInterruptMode,
                DeviceExtension->bDevicePCIInterruptShared,
                DeviceExtension->DevicePCIInterruptAffinity,
                FALSE
                );

    if (!NT_SUCCESS(status))
    {
        return status;
    }


    //*****************************************************************
    //*****************************************************************
    // TODO: Reset and initialize your hardware
    //*****************************************************************
    //*****************************************************************

    //*****************************************************************
    //*****************************************************************
    // TODO: enable interrupts
    //*****************************************************************
    //*****************************************************************

    DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIFreeResources
//      This routine returns all the resources acquired during
//      device startup.  Here we disconnect from any interrupt, 
//      unmap any I/O ports that are mapped in StartDevice, and 
//      disable any device interfaces or symbolic links.  
//      Before disconnecting an interrupt, be sure the device can
//      no longer generate interrupts.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      NT status code
//
NTSTATUS DevicePCIFreeResources(
    IN  PDEVICEPCI_DEVICE_EXTENSION DeviceExtension
    )
{
    NTSTATUS    status;

    DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"++");

    // If there is an interrupt object, disconnect the interrupt
    if (DeviceExtension->DevicePCIInterruptObject != NULL)
    {
        // Disconnect our ISR from the interrupt resource
        IoDisconnectInterrupt(DeviceExtension->DevicePCIInterruptObject);
        DeviceExtension->DevicePCIInterruptObject = NULL;
    }

    if (DeviceExtension->Resource_portBase != NULL)
    {
        if (DeviceExtension->bResource_portMapped)
        {
            MmUnmapIoSpace(DeviceExtension->Resource_portBase, DeviceExtension->Resource_portLength);
        }

        DeviceExtension->Resource_portBase = NULL;
    }

    DevicePCIDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"--");

    return STATUS_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCICreateDispatch
//      Dispatch routine for IRP_MJ_CREATE requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to the device object for our device
//
//      IN  Irp
//              the create IRP
//
//  Return Value:
//      NT status code.
//
NTSTATUS DevicePCICreateDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PDEVICEPCI_DEVICE_EXTENSION    deviceExtension;
    NTSTATUS                status;

    DevicePCIDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__": IRP %p", Irp);

    deviceExtension = (PDEVICEPCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    if (!DevicePCIAcquireRemoveLock(deviceExtension))
    {
        status = STATUS_NO_SUCH_DEVICE;

        Irp->IoStatus.Status = status;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);

        DevicePCIDebugPrint(DBG_CREATECLOSE, DBG_WARN, __FUNCTION__"$$--. IRP %p, STATUS %x", Irp, status);

        return status;
    }

    // increment open count
    InterlockedIncrement(&deviceExtension->OpenHandleCount);

    status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    DevicePCIReleaseRemoveLock(deviceExtension);

    DevicePCIDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"--. IRP %p, STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCICloseDispatch
//      Dispatch routine for IRP_MJ_CLOSE requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to the device object for our device
//
//      IN  Irp
//              the close IRP
//
//  Return Value:
//      NT status code.
//
NTSTATUS DevicePCICloseDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PDEVICEPCI_DEVICE_EXTENSION    deviceExtension;
    NTSTATUS                status;

    DevicePCIDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    deviceExtension = (PDEVICEPCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = status;
    IoCompleteRequest (Irp, IO_NO_INCREMENT);

    InterlockedDecrement(&deviceExtension->OpenHandleCount);

    DevicePCIDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"--. IRP %p, STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCICleanupDispatch
//      Dispatch routine for IRP_MJ_CLEANUP requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to the device object for our device
//
//      IN  Irp
//              the cleanup IRP
//
//  Return Value:
//      NT status code.
//
NTSTATUS DevicePCICleanupDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PDEVICEPCI_DEVICE_EXTENSION    deviceExtension;
    NTSTATUS                status;
    PIO_STACK_LOCATION      irpStack;

    DevicePCIDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    deviceExtension = (PDEVICEPCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    irpStack = IoGetCurrentIrpStackLocation(Irp);
    DevicePCIFlushQueues(deviceExtension, irpStack->FileObject);

    status = STATUS_SUCCESS;

    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    DevicePCIDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIUnload
//      Driver Unload routine.
//
//  Arguments:
//      IN  DriverObject
//              pointer to the driver object
//
//  Return Value:
//      none
//
VOID DevicePCIUnload(
    IN  PDRIVER_OBJECT  DriverObject
    )
{
    DevicePCIDebugPrint(DBG_UNLOAD, DBG_TRACE, __FUNCTION__"++");

    // The device object(s) should be NULL now
    // (since we unload, all the devices objects associated with this
    // driver must be deleted.
    ASSERT(DriverObject->DeviceObject == NULL);

    // We should not be unloaded until all the devices we control
    // have been removed from our queue.

    // release memory block allocated for registry path
    if (g_Data.RegistryPath.Buffer != NULL)
    {
        ExFreePool(g_Data.RegistryPath.Buffer);
        g_Data.RegistryPath.Buffer = NULL;
    }

    DevicePCIDebugPrint(DBG_UNLOAD, DBG_TRACE, __FUNCTION__"--");

#ifdef DEVICEPCI_WMI_TRACE
    WPP_CLEANUP(DriverObject);
#endif

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIIsStoppable
//      This routine determines whether the device can be safely stopped. 
//      In our particular case, we'll assume we can always stop the device.
//      A device might fail the request if it doesn't have a queue for the
//      requests that might come or if it was notified that it is in the 
//      paging path.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      returns TRUE if the device is stoppable, FALSE otherwise
//
BOOLEAN DevicePCIIsStoppable(
    IN  PDEVICEPCI_DEVICE_EXTENSION    DeviceExtension
    )
{
    return TRUE;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIIsRemovable
//      This routine determines whether the device can be safely removed. 
//      A device shouldn't be removed if, for example, it has open handles or
//      removing the device could result in losing data (plus the reasons
//      mentioned in the DevicePCIIsStoppable function comments). The PnP manager on 
//      Windows 2000 fails on its own any attempt to remove, if there any 
//      open handles to the device.  However on Win9x, the driver must keep 
//      count of open handles and fail QueryRemove if there are any open 
//      handles.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      returns TRUE if the device is stoppable, FALSE otherwise
//
BOOLEAN DevicePCIIsRemovable(
    IN  PDEVICEPCI_DEVICE_EXTENSION    DeviceExtension
    )
{
    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCISubmitIrpSyncComplete
//      Completion routine for sync IRP requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to our device object
//
//      IN  Irp
//              pointer to the PnP IRP
//
//      IN  Context
//              our event used to signal IRP completion
//
//  Return Value:
//      STATUS_MORE_PROCESSING_REQUIRED
//
NTSTATUS  DevicePCISubmitIrpSyncComplete(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp,
    IN  PVOID           Context
    )
{
    PKEVENT event = (PKEVENT)Context;

    // If the lower driver didn't return STATUS_PENDING, we don't need to 
    // set the event because we won't be waiting on it. 
    if (Irp->PendingReturned) 
    {
        KeSetEvent(event, IO_NO_INCREMENT, FALSE);
    }

    return STATUS_MORE_PROCESSING_REQUIRED;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCISubmitIrpSync
//      Sends the given IRP down the stack to the next lower driver and 
//      waits in a synchronous fashion for the IRP to complete
//
//  Arguments:
//      IN  DeviceObject
//              Pointer to device object for our device
//
//      IN  Irp
//              IRP to send down
//
//  Return Value:
//      NT status code
//
NTSTATUS DevicePCISubmitIrpSync(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    KEVENT   event;
    NTSTATUS status;

    KeInitializeEvent(&event, NotificationEvent, FALSE);

    IoCopyCurrentIrpStackLocationToNext(Irp);

    IoSetCompletionRoutine(
        Irp,
        DevicePCISubmitIrpSyncComplete,
        &event,
        TRUE,
        TRUE,
        TRUE
        );

    status = IoCallDriver(DeviceObject, Irp);

    // Wait for lower drivers to be done with the Irp.
    // Important thing to note here is when you allocate
    // memory for an event in the stack you must do a
    // KernelMode wait instead of UserMode to prevent
    // the stack from getting paged out.

    if (status == STATUS_PENDING) 
    {
       KeWaitForSingleObject(
           &event,
           Executive,
           KernelMode,
           FALSE,
           NULL
           );

       status = Irp->IoStatus.Status;
    }

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIAcquireRemoveLock
//      Acquires remove lock.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      FALSE if remove device pending, TRUE otherwise.
//
BOOLEAN DevicePCIAcquireRemoveLock(
    IN  PDEVICEPCI_DEVICE_EXTENSION    DeviceExtension
    )
{
    LONG    count;

    count = InterlockedIncrement(&DeviceExtension->RemoveCount);

    ASSERT(count > 0);

    if (DeviceExtension->PnpState == PnpStateRemoved)
    {
        DevicePCIReleaseRemoveLock(DeviceExtension);

        return FALSE;
    }

    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIReleaseRemoveLock
//      Releases remove lock.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      None.
//
VOID DevicePCIReleaseRemoveLock(
    IN  PDEVICEPCI_DEVICE_EXTENSION    DeviceExtension
    )
{
    LONG    count;

    count = InterlockedDecrement(&DeviceExtension->RemoveCount);

    ASSERT(count >= 0);

    if (count == 0)
    {
        KeSetEvent(&DeviceExtension->RemoveEvent, IO_NO_INCREMENT, FALSE);
    }

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIWaitForSafeRemove
//      Waits for all remove locks to be released
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      None.
//
//  Comment:
//      This routine should be called with no remove locks held 
//      by the calling thread
//
VOID DevicePCIWaitForSafeRemove(
    IN  PDEVICEPCI_DEVICE_EXTENSION    DeviceExtension
    )
{
    DeviceExtension->PnpState = PnpStateRemoved;
    DevicePCIReleaseRemoveLock(DeviceExtension);

    KeWaitForSingleObject(
        &DeviceExtension->RemoveEvent, 
        Executive, 
        KernelMode, 
        FALSE, 
        NULL
        );

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIStallQueues
//      Pauses all of the queues, and waits for them to get to paused state.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      none
//
VOID DevicePCIStallQueues(
    IN  PDEVICEPCI_DEVICE_EXTENSION DeviceExtension
    )
{
    ULONG   index;

    DevicePCIDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++");

    // stall IRP processing
    DevicePCILockIo(&DeviceExtension->IoLock);

    DevicePCIWaitForStopIo(&DeviceExtension->IoLock);

    DevicePCIDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--");

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIRestartQueues
//      Restarts all paused queues.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      none
//
VOID DevicePCIRestartQueues(
    IN  PDEVICEPCI_DEVICE_EXTENSION   DeviceExtension
    )
{
    ULONG   index;

    DevicePCIDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++");

    DevicePCIUnlockIo(&DeviceExtension->IoLock);

    DevicePCIDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--");

    return;
}
///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIFlushQueues
//      Flush oustanding IRPs for closed file object.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  FileObject
//              about to be closed file object
//
//  Return Value:
//      none
//
VOID DevicePCIFlushQueues(
    IN  PDEVICEPCI_DEVICE_EXTENSION   DeviceExtension,
    IN  PFILE_OBJECT            FileObject
    )
{
    ULONG   index;

    DevicePCIDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++");

    DevicePCIFlushPendingIo(&DeviceExtension->IoLock, FileObject);

    DevicePCIDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--");

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DevicePCIAccessConfigSpace
//      Read or Write PCI configuration space
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  IsRead
//              TRUE for reads, FALSE for writes
//
//      IN  Buffer
//              Buffer to read to or write from
//
//      IN  Offset
//              starting offset in PCI configuration space
//
//      IN  Length
//              Buffer size
//
//      IN  ReadWrittenSize
//              count of bytes read or written
//
//  Return Value:
//      Status
//
NTSTATUS DevicePCIAccessConfigSpace(
    IN  PDEVICEPCI_DEVICE_EXTENSION DeviceExtension,
    IN  BOOLEAN                 IsRead,
    IN OUT PVOID                Buffer,
    IN  ULONG                   Offset,
    IN  ULONG                   Length,
    OUT PULONG                  ReadWrittenSize
    )
{
    NTSTATUS            status;
    PDEVICE_OBJECT      topDeviceObject;
    PIRP                irp;
    PIO_STACK_LOCATION  irpStack;
    KEVENT              event;
    IO_STATUS_BLOCK     ioStatus;

    DevicePCIDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++");

    KeInitializeEvent(&event, NotificationEvent, FALSE);

    topDeviceObject = IoGetAttachedDeviceReference(DeviceExtension->DeviceObject);

    irp = IoBuildSynchronousFsdRequest(
            IRP_MJ_PNP,
            topDeviceObject,
            NULL,
            0,
            NULL,
            &event,
            &ioStatus
            );

    if (irp != NULL)
    {
        irpStack = IoGetNextIrpStackLocation(irp);

        irpStack->MinorFunction = IsRead ? IRP_MN_READ_CONFIG : IRP_MN_WRITE_CONFIG;

        irpStack->Parameters.ReadWriteConfig.WhichSpace = PCI_WHICHSPACE_CONFIG;
        irpStack->Parameters.ReadWriteConfig.Buffer = Buffer;
        irpStack->Parameters.ReadWriteConfig.Offset = Offset;
        irpStack->Parameters.ReadWriteConfig.Length = Length;
        
        irp->IoStatus.Status = STATUS_NOT_SUPPORTED;
        
        status = IoCallDriver(topDeviceObject, irp);
        if (status == STATUS_PENDING)
        {
            KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
            status = ioStatus.Status;
        }

        *ReadWrittenSize = (ULONG)ioStatus.Information;
    }
    else
    {
        status = STATUS_INSUFFICIENT_RESOURCES;
    }

    ObDereferenceObject(topDeviceObject);

    DevicePCIDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--. STATUS %x", status);

    return status;
}
